/*
 * kernel/power/proc.c
 *
 * Copyright (C) 2002-2006 Nigel Cunningham <nigel@suspend2.net>
 *
 * This file is released under the GPLv2.
 *
 * This file contains support for proc entries for tuning Suspend2.
 *
 * We have a generic handler that deals with the most common cases, and
 * hooks for special handlers to use.
 */

#include <linux/suspend.h>
#include <linux/module.h>
#include <asm/uaccess.h>

#include "proc.h"
#include "suspend2.h"
#include "storage.h"

static int suspend_proc_initialised = 0;

static struct list_head suspend_proc_entries;
static struct proc_dir_entry *suspend_dir;
static struct suspend_proc_data proc_params[];

extern void __suspend_try_resume(void);
extern void suspend_main(void);

/* suspend_read_proc
 *
 * Generic handling for reading the contents of bits, integers,
 * unsigned longs and strings.
 */
static int suspend_read_proc(char *page, char **start, off_t off, int count,
		int *eof, void *data)
{
	int len = 0;
	struct suspend_proc_data *proc_data = (struct suspend_proc_data *) data;

	if (suspend_start_anything(0))
		return -EBUSY;
	
	if (proc_data->needs_storage_manager & 1)
		suspend_prepare_usm();
	
	switch (proc_data->type) {
		case SUSPEND_PROC_DATA_CUSTOM:
			if (proc_data->data.special.read_proc) {
				read_proc_t *read_proc = proc_data->data.special.read_proc;
				len = read_proc(page, start, off, count, eof, data);
			} else
				len = 0;
			break;
		case SUSPEND_PROC_DATA_BIT:
			len = sprintf(page, "%d\n", 
				-test_bit(proc_data->data.bit.bit,
					proc_data->data.bit.bit_vector));
			break;
		case SUSPEND_PROC_DATA_INTEGER:
			{
				int *variable = proc_data->data.integer.variable;
				len = sprintf(page, "%d\n", *variable);
				break;
			}
		case SUSPEND_PROC_DATA_LONG:
			{
				long *variable = proc_data->data.a_long.variable;
				len = sprintf(page, "%ld\n", *variable);
				break;
			}
		case SUSPEND_PROC_DATA_UL:
			{
				long *variable = proc_data->data.ul.variable;
				len = sprintf(page, "%lu\n", *variable);
				break;
			}
		case SUSPEND_PROC_DATA_STRING:
			{
				char *variable = proc_data->data.string.variable;
				len = sprintf(page, "%s\n", variable);
				break;
			}
	}
	/* Side effect routine? */
	if (proc_data->read_proc)
		proc_data->read_proc();
	
	if (len <= count)
		*eof = 1;
	
	if (proc_data->needs_storage_manager & 1)
		suspend_cleanup_usm();

	suspend_finish_anything(0);
	
	return len;
}

/* suspend_write_proc
 *
 * Generic routine for handling writing to files representing
 * bits, integers and unsigned longs.
 */

static int suspend_write_proc(struct file *file, const char *buffer,
		unsigned long count, void *data)
{
	struct suspend_proc_data *proc_data = (struct suspend_proc_data *) data;
	char *my_buf = (char *) get_zeroed_page(GFP_ATOMIC);
	int result = count, assigned_temp_buffer = 0;

	if (!my_buf)
		return -ENOMEM;

	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(my_buf, buffer, count))
		return -EFAULT;
	
	if (suspend_start_anything(proc_data == &proc_params[0]))
		return -EBUSY;

	my_buf[count] = 0;

	if (proc_data->needs_storage_manager & 2)
		suspend_prepare_usm();

	switch (proc_data->type) {
		case SUSPEND_PROC_DATA_CUSTOM:
			if (proc_data->data.special.write_proc) {
				write_proc_t *write_proc = proc_data->data.special.write_proc;
				result = write_proc(file, buffer, count, data);
			}
			break;
		case SUSPEND_PROC_DATA_BIT:
			{
			int value = simple_strtoul(my_buf, NULL, 0);
			if (value)
				set_bit(proc_data->data.bit.bit, 
					(proc_data->data.bit.bit_vector));
			else
				clear_bit(proc_data->data.bit.bit,
					(proc_data->data.bit.bit_vector));
			}
			break;
		case SUSPEND_PROC_DATA_INTEGER:
			{
				int *variable = proc_data->data.integer.variable;
				int minimum = proc_data->data.integer.minimum;
				int maximum = proc_data->data.integer.maximum;
				*variable = simple_strtol(my_buf, NULL, 0);
				if (((*variable) < minimum))
					*variable = minimum;

				if (((*variable) > maximum))
					*variable = maximum;
				break;
			}
		case SUSPEND_PROC_DATA_LONG:
			{
				long *variable = proc_data->data.a_long.variable;
				long minimum = proc_data->data.a_long.minimum;
				long maximum = proc_data->data.a_long.maximum;
				*variable = simple_strtol(my_buf, NULL, 0);
				
				if (minimum && ((*variable) < minimum))
					*variable = minimum;

				if (maximum && ((*variable) > maximum))
					*variable = maximum;
				break;
			}
		case SUSPEND_PROC_DATA_UL:
			{
				unsigned long *variable = proc_data->data.ul.variable;
				unsigned long minimum = proc_data->data.ul.minimum;
				unsigned long maximum = proc_data->data.ul.maximum;
				*variable = simple_strtoul(my_buf, NULL, 0);
				
				if (minimum && ((*variable) < minimum))
					*variable = minimum;

				if (maximum && ((*variable) > maximum))
					*variable = maximum;
				break;
			}
			break;
		case SUSPEND_PROC_DATA_STRING:
			{
				int copy_len = count;
				char *variable =
					proc_data->data.string.variable;

				if (proc_data->data.string.max_length &&
				    (copy_len > proc_data->data.string.max_length))
					copy_len = proc_data->data.string.max_length;

				if (!variable) {
					proc_data->data.string.variable =
						variable = (char *) get_zeroed_page(GFP_ATOMIC);
					assigned_temp_buffer = 1;
				}
				strncpy(variable, my_buf, copy_len);
				if ((copy_len) &&
					 (my_buf[copy_len - 1] == '\n'))
					variable[count - 1] = 0;
				variable[count] = 0;
			}
			break;
	}
	free_page((unsigned long) my_buf);
	/* Side effect routine? */
	if (proc_data->write_proc)
		proc_data->write_proc();

	/* Free temporary buffers */
	if (assigned_temp_buffer) {
		free_page((unsigned long) proc_data->data.string.variable);
		proc_data->data.string.variable = NULL;
	}

	if (proc_data->needs_storage_manager & 2)
		suspend_cleanup_usm();

	suspend_finish_anything(proc_data == &proc_params[0]);
	
	return result;
}

/* Non-module proc entries.
 *
 * This array contains entries that are automatically registered at
 * boot. Modules and the console code register their own entries separately.
 *
 * NB: If you move do_suspend, change suspend_write_proc's test so that
 * suspend_start_anything still gets a 1 when the user echos > do_suspend!
 */

static struct suspend_proc_data proc_params[] = {
	{ .filename			= "do_suspend",
	  .permissions			= PROC_WRITEONLY,
	  .type				= SUSPEND_PROC_DATA_CUSTOM,
	  .write_proc			= suspend_main,
	  .needs_storage_manager	= 2,
	},

	{ .filename			= "do_resume",
	  .permissions			= PROC_WRITEONLY,
	  .type				= SUSPEND_PROC_DATA_CUSTOM,
	  .write_proc			= __suspend_try_resume,
	  .needs_storage_manager	= 2,
	},
};
       
/* suspend_initialise_proc
 *
 * Initialise the /proc/suspend2 directory.
 */

static void suspend_initialise_proc(void)
{
	int i;
	int numfiles = sizeof(proc_params) / sizeof(struct suspend_proc_data);
	
	if (suspend_proc_initialised)
		return;

	suspend_dir = proc_mkdir("suspend2", NULL);
	
	BUG_ON(!suspend_dir);

	INIT_LIST_HEAD(&suspend_proc_entries);

	suspend_proc_initialised = 1;

	for (i=0; i< numfiles; i++)
		suspend_register_procfile(&proc_params[i]);
}

/* suspend_register_procfile
 *
 * Helper for registering a new /proc/suspend2 entry.
 */

struct proc_dir_entry *suspend_register_procfile(
		struct suspend_proc_data *suspend_proc_data)
{
	struct proc_dir_entry *new_entry;
	
	if (!suspend_proc_initialised)
		suspend_initialise_proc();

	new_entry = create_proc_entry(
			suspend_proc_data->filename,
			suspend_proc_data->permissions, 
			suspend_dir);
	if (new_entry) {
		list_add_tail(&suspend_proc_data->proc_data_list, &suspend_proc_entries);
		new_entry->read_proc = suspend_read_proc;
		new_entry->write_proc = suspend_write_proc;
		new_entry->data = suspend_proc_data;
	} else {
		printk("Error! create_proc_entry returned NULL.\n");
		INIT_LIST_HEAD(&suspend_proc_data->proc_data_list);
	}
	return new_entry;
}

/* suspend_unregister_procfile
 *
 * Helper for removing unwanted /proc/suspend2 entries.
 *
 */
void suspend_unregister_procfile(struct suspend_proc_data *suspend_proc_data)
{
	if (list_empty(&suspend_proc_data->proc_data_list))
		return;

	remove_proc_entry(
		suspend_proc_data->filename,
		suspend_dir);
	list_del(&suspend_proc_data->proc_data_list);
}
